


function [Ts times ind_obj ind_gn mask]=pbhs_solve(T0,dur,DX,rho,C,K,Wb,Cb,Tb,Q,options)
% function [Ts times ind_obj ind_gn mask]=pbhs_solve(T0,dur,DX,rho,C,K,Wb,Cb,Tb,Q,options)
%
% INPUTS:
% T0: map of initial temperature (K)
% dur: Duration (s)
% DX: 1*3 vector of pixel dimensions ([dx dy dz]) in m
% rho: Density map (kg/m^3)
% C: Heat capacity map (J/kg/K)
% K: Thermal cond. map (W/m/K)
% Wb: Perfusion map (kg/m^3/s)
% Cb: Heat capacity of blood (J/kg/K, good default value is 3617 J/kg/K)
% Tb: Blood temperature (K)
% Q: Bioheat generation density (W/m^3, == SAR * density)
% 
% OPTIONS:
% KeepAllTempMaps: Keep all temperature maps is ==1
% DisplayTemp: Display temperature maps as algo pregresses is ==1
% UsePreconditioner: Use LU if ==1 (default ==0)
% LUPreconDropTol: LU precon tolerance (default ==1e-2)
% tol_lsqr: Convergence tolerance for each time step (default ==1e-6)
% niter_lsqr: Max # of iteration per LSQR step (default == 1e6)
% BoundaryConditionType (1: Dirichlet, 2: convection): Default is Dirichlet
% BoundaryTemperature: Room temperature (default ==25)
% TimeStep: Time step (s, default is computed automatically)
% NeumannConvCoef: COnvection coeficient, only used if BoundaryConditionType==2 (W/m^2/K)
% T_inf: Temperaure of convective reservoir, only used if BoundaryConditionType==2 (K)
%
% OUTPUTS:
% Ts: Temperature maps (K)
% times: Time points (s)
% ind_obj: Indices of object
% ind_gn: Indices of ghost nodes
% mask: 1 (inside object), 2 (boundary pixels) or 3 (ghost nodes)
% 

addpath('F:/Computation/MATLAB/Library/sparsesubaccess');

% set up options
options=decode_options(options);

% create ghost nodes
mask=double(rho>100);
ind_b_global=pbhs_detect_pixels_boundary(mask);  % global indices
ind_gn_global=pbhs_create_ghost_nodes(mask,ind_b_global);  % global indices

mask(ind_b_global)=2;  % 1 is within the object, 2 is for boundary pixels and 3 is for ghost nodes
mask(ind_gn_global)=3;

ind=find(mask>0);
npix=size(ind,1);  % total number of pixels, including ghost nodes

ind_b=find(mask(ind)==2);  % local indices
ind_gn=find(mask(ind)==3);  % local indices

ind_obj_global=find(mask==1 | mask==2);
ind_obj=find(mask(ind)==1 | mask(ind)==2);

% constant diagonal terms
AA=sparse(npix,npix);
BB=sparse(npix,npix);
CC=sparse(npix,npix);

AA=setsparse( AA,[1:npix]',[1:npix]',1./(rho(ind).*C(ind))          );
BB=setsparse( BB,[1:npix]',[1:npix]',Cb*Wb(ind)./(rho(ind).*C(ind)) );
CC=setsparse( CC,[1:npix]',[1:npix]',Q(ind)./(rho(ind).*C(ind))     );

AA=setsparse( AA,ind_gn,ind_gn,0 );
BB=setsparse( BB,ind_gn,ind_gn,0 );
CC=setsparse( CC,ind_gn,ind_gn,0 );

% compute time step
deltat_neumann=2 ./ ( diag(BB) + 4*diag(AA)/min(DX)^2 );
if options.TimeStep>0
    deltat=options.TimeStep;
    fprintf('Using user-defined time step of %f seconds (Von Neumann criterion is %f)\n',deltat,min(deltat_neumann(:)));
else    
    ind2 = find( rho(ind)>100 | mask(ind)==3 );  % exclude internal air from the Neuman step calculation, otherwise it takes forever    
                                                % but don't forget to include the ghost nodes (mask==3)!!
    deltat=min(deltat_neumann(ind2));
    fprintf('Using Von Neumann time step of %f seconds\n',deltat);
end
ntimes=ceil( dur/deltat );

% create Fourier operator
LAP=pbhs_create_laplacian_operator_v3(mask,DX,K);

% create system matrices for LHS and RHS @ each time step
fprintf('Creating system matrices\n');
EYE=sparse(npix,npix);
EYE=setsparse( EYE,[1:npix]',[1:npix]',1 );
A=EYE - deltat/2*( AA*LAP - BB );
B=EYE + deltat/2*( AA*LAP - BB );

% modify A and B so as to enforce Dirichlet at the ghost nodes
% A(ind_gn,:)=0; A(:,ind_gn)=0; A(ind_gn_diag)=1;
% B(ind_gn,:)=0; B(:,ind_gn)=0; B(ind_gn_diag)=1;

% compute preconditioner
if options.UsePreconditioner==1
    fprintf('Computing incomplete LU preconditioner using drop tolerance = %f\n',options.LUPreconDropTol);
    [L U]=ilu(A,struct('type','ilutp','droptol',options.LUPreconDropTol));
end

fprintf('Iterating in time\n');
sbuf=[];
T_to_disp=zeros(size(mask));
T0=T0.*double( mask==1 | mask==2 );

if options.DisplayTemp==1
    f=figure;
end

if options.KeepAllTempMaps==1
    tmp=zeros(size(mask));
    Ts=zeros( size(mask,1),size(mask,2),size(mask,3),ntimes );
    times=[1:ntimes]'*deltat;
else
    Ts=zeros( size(mask) );
    times=ntimes*deltat;
end

for i=1:ntimes
    
    if i==1
        Tprev=T0(ind);
    end
    
    Tprev=pbhs_update_ghost_nodes(Tprev,options,ind_gn);
    
    fprintf('\tTime step %d/%d: Creating right hand side...  ',i,ntimes);
    b=B*Tprev + diag(CC)*deltat + diag(BB)*Tb*deltat;
    
    fprintf('Solving sparse linear system using LSQR...  ');
    tic;
    if options.UsePreconditioner==1
        [Tprev flag relres iter]=lsqr(A,b,options.tol_lsqr,options.niters_lsqr,L,U,Tprev);
    else
        [Tprev flag relres iter]=lsqr(A,b,options.tol_lsqr,options.niters_lsqr,[],[],Tprev);
    end
    t2=toc;
       
    fprintf('Solve time=%f seconds   flag=%d   rel. residual=%3.3e   # of LSQR iter.=%d\n',t2,flag,relres,iter(1));
    
    if options.DisplayTemp==1
        T_to_disp(ind)=Tprev;
        subplot(2,2,1); imagesc(squeeze(T_to_disp(:,:,ceil(end/2)))); axis image; axis off; colormap(hot); colorbar;
        title( sprintf('T (%f min.) [degrees C]',deltat*i/60) );
        subplot(2,2,2); imagesc(squeeze(T_to_disp(:,ceil(end/2),:))); axis image; axis off; colormap(hot); colorbar;        
        subplot(2,2,3); imagesc(squeeze(T_to_disp(ceil(end/2),:,:))); axis image; axis off; colormap(hot); colorbar;
        subplot(2,2,4); imagesc(squeeze(max(T_to_disp,[],1))); axis image; axis off; colormap(hot); colorbar; title('MIP');
        drawnow;
    end
    
    if options.KeepAllTempMaps==1
        tmp(ind_obj_global)=Tprev(ind_obj);
        Ts(:,:,:,i)=tmp;
    end
    
end

if options.KeepAllTempMaps==0
    Ts(ind_obj_global)=Tprev(ind_obj);
end





function options2=decode_options(options)

options2=options;

if isfield(options2,'UsePreconditioner') && options2.UsePreconditioner==1
    if isempty(options2.LUPreconDropTol) || options2.LUPreconDropTol<=0
        options2.LUPreconDropTol=1e-2;
    end
else
    options2.UsePreconditioner=0;
end

if ~isfield(options2,'tol_lsqr') || isempty(options2.tol_lsqr) || options2.tol_lsqr<=0
    options2.tol_lsqr=1e-6;
end

if ~isfield(options2,'niters_lsqr') || isempty(options2.niters_lsqr) || options2.niters_lsqr<=0
    options2.niters_lsqr=1e6;
end

if ~isfield(options2,'BoundaryConditionType') || isempty(options2.BoundaryConditionType)
    options2.BoundaryConditionType=1;
end

if options2.BoundaryConditionType==2 && ~isfield(options2,'NeumannConvCoef')
    options.NeumannConvCoef=10.5;  % W/m^2/K
end

if options2.BoundaryConditionType==2 && ~isfield(options2,'T_inf')
    options.T_inf=25;
end

if ~isfield(options2,'BoundaryTemperature') || isempty(options2.BoundaryTemperature)
    options2.BoundaryTemperature=25;
end

if ~isfield(options2,'TimeStep') || isempty(options2.TimeStep)
    options2.TimeStep=-1;
end

















































