


function T_SS = pbhs_solve(DX,rho,C,K,Wb,Cb,Tb,Q,options)
% function T_SS = pbhs_solve(DX,rho,C,K,Wb,Cb,Tb,Q,options)
%
% INPUTS:
% DX: 1*3 vector of pixel dimensions ([dx dy dz]) in m
% rho: Density map (kg/m^3)
% C: Heat capacity map (J/kg/K)
% K: Thermal cond. map (W/m/K)
% Wb: Perfusion map (kg/m^3/s)
% Cb: Heat capacity of blood (J/kg/K, good default value is 3617 J/kg/K)
% Tb: Blood temperature (K)
% Q: Bioheat generation density (W/m^3, == SAR * density)
% 
% OPTIONS:
% KeepAllTempMaps: Keep all temperature maps is ==1
% DisplayTemp: Display temperature maps as algo pregresses is ==1
% UsePreconditioner: Use LU if ==1 (default ==0)
% LUPreconDropTol: LU precon tolerance (default ==1e-2)
% tol_lsqr: Convergence tolerance for each time step (default ==1e-6)
% niters_lsqr: Max # of iteration per LSQR step (default == 1e6)
% BoundaryConditionType (1: Dirichlet, 2: convection): Default is Dirichlet
% BoundaryTemperature: Room temperature (default ==25)
% TimeStep: Time step (s, default is computed automatically)
% NeumannConvCoef: COnvection coeficient, only used if BoundaryConditionType==2 (W/m^2/K)
% T_inf: Temperaure of convective reservoir, only used if BoundaryConditionType==2 (K)
%
% OUTPUTS:
% Ts: Temperature maps (K)
% times: Time points (s)
% ind_obj: Indices of object
% ind_gn: Indices of ghost nodes
% mask: 1 (inside object), 2 (boundary pixels) or 3 (ghost nodes)
% 

addpath('F:/Computation/MATLAB/Library/sparsesubaccess');

% set up options
options=decode_options(options);

% create ghost nodes
mask=double(rho>100);
ind=find(mask>0);
npix=size(ind,1);  % total number of pixels, including ghost nodes

% constant diagonal term
BB=sparse(npix,npix);
BB=setsparse( BB,[1:npix]',[1:npix]',Cb*Wb(ind) );

% create Fourier operator
LAP=pbhs_create_laplacian_operator_v3(mask,DX,K);

% create system matrices for LHS and RHS @ each time step
fprintf('Creating LHS and RHS\n');
A = LAP - BB;
b = -Q(ind) - Wb(ind)*Tb*Cb;
    
fprintf('Solving sparse linear system using LSQR... \n');
[T flag relres iter]=lsqr_(A,b,options.tol_lsqr,options.niters_lsqr);

T_SS = zeros(size(mask));
T_SS(ind)=T;

if options.DisplayTemp==1
    subplot(1,3,1); imagesc(squeeze(T_SS(:,:,ceil(end/2)))); axis image; axis off; colormap(hot); colorbar;
    subplot(1,3,2); imagesc(squeeze(T_SS(:,ceil(end/2),:))); axis image; axis off; colormap(hot); colorbar;
    title('Steady-state T [degrees C]');
    subplot(1,3,3); imagesc(squeeze(T_SS(ceil(end/2),:,:))); axis image; axis off; colormap(hot); colorbar;
    drawnow;
end
   







function options2=decode_options(options)

options2=options;

if isfield(options2,'UsePreconditioner') && options2.UsePreconditioner==1
    if isempty(options2.LUPreconDropTol) || options2.LUPreconDropTol<=0
        options2.LUPreconDropTol=1e-2;
    end
else
    options2.UsePreconditioner=0;
end

if ~isfield(options2,'tol_lsqr') || isempty(options2.tol_lsqr) || options2.tol_lsqr<=0
    options2.tol_lsqr=1e-6;
end

if ~isfield(options2,'niters_lsqr') || isempty(options2.niters_lsqr) || options2.niters_lsqr<=0
    options2.niters_lsqr=1e6;
end

if ~isfield(options2,'BoundaryConditionType') || isempty(options2.BoundaryConditionType)
    options2.BoundaryConditionType=1;
end

if options2.BoundaryConditionType==2 && ~isfield(options2,'NeumannConvCoef')
    options.NeumannConvCoef=10.5;  % W/m^2/K
end

if options2.BoundaryConditionType==2 && ~isfield(options2,'T_inf')
    options.T_inf=25;
end

if ~isfield(options2,'BoundaryTemperature') || isempty(options2.BoundaryTemperature)
    options2.BoundaryTemperature=25;
end

if ~isfield(options2,'TimeStep') || isempty(options2.TimeStep)
    options2.TimeStep=-1;
end

















































